#!/usr/bin/env python

################################################ IMPORTS ########################################
import os
import sys
import datetime
import subprocess
from collections import OrderedDict
try:
    from pyfaidx import Fasta
except ImportError as e:
    sys.exit("ERROR: Have you activated the conda environment? I got the following error: "+str(e)+"\n aborting.")
import pandas as pd
import pyBigWig
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import random
import argparse
from scipy import stats
import glob
import multiprocessing
import h2o
from h2o.estimators.glm import H2OGeneralizedLinearEstimator
from h2o.estimators.random_forest import H2ORandomForestEstimator
from h2o.grid.grid_search import H2OGridSearch
from h2o.automl import H2OAutoML
from sklearn.preprocessing import PolynomialFeatures
from collections import defaultdict
from Bio import Seq

################################################ ARGUMENTS #####################################
parser = argparse.ArgumentParser(description="LIQUORICE: A tool for bias correction in cfDNA WGS datasets",
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)
required_keyword_args = parser.add_argument_group('required named arguments')

required_keyword_args.add_argument('--samplename', help='Name of the sample that is being processed. '
                                                              'This will be used for ouput plots and the names of '
                                                              'directories and log files', required=True)
required_keyword_args.add_argument('--bedpath', help='Path to the BED file that contains the regions that '
                                                           'should be analyzed. Should be in hg38. '
                                                           'Should be in the format chromName\\tstart\\tend',
                                   required=True)
required_keyword_args.add_argument('--bigwigpath', help='Path to the BigWig file that contains the coverage information'
                                                        ' of the sample. Should probably have a small binsize, '
                                                        'we are working with bamCoverage with bs=1'
                                                        '', required=True)
required_keyword_args.add_argument('--fraglenpath', help='Path to the output .txt file of Picard\'s '
                                                          'CollectInsertSizeMetrics tool.' ,required=True)
required_keyword_args.add_argument('--cnvpath', help='Path to ichorCNA\s <samplename>.correctedDepth.txt file. Required'
                                                     'for correction of CNV biases' ,required=True)

required_keyword_args.add_argument('--hg38path', help='Path to a hg38.fa file' ,required=True)
required_keyword_args.add_argument('--chromsizespath', help='Path to a chrom.sizes file for hg38' ,required=True)
required_keyword_args.add_argument('--mapppath', help='Path to a bigWig file that contains (forward) mappability'
                                                            ' values for every base in the reference genome. '
                                                            'Can be calculated with gem-mappability for the appropriate '
                                                            'read length' ,required=True)


parser.add_argument('--binsize', help="Bin size that is to be used for the flanking regions. All bins except"
                                           " the core region,"
                                           "specified in --bedpath (which will be splitted differently), will have "
                                           "this binsize. "
                                           "Bin size is important for the resolution of the output plots & data, and "
                                           "for the bias model itself (default: 180)", default=180, type=int)

parser.add_argument('--extendto', help="Size of the flanking region, in bp. Must be devidable by bin size "
                                       "The analyzed regions are extended by "
                                            "this value, both upstream and downstream. Example:"
                                            "core region=150 bp, extendto=1000 --> length of analyzed region in total: "
                                            "2150 bp (default: 18000) ", default=18000, type=int)
parser.add_argument('--dontcorrectCNVs', help="If set, do not correct for CNV biases. You can set --cnvpath to 'None' if"
                                              "this option is used. If a file is given, it will be ignored."
                    , action="store_true")

parser.add_argument('--readlength', help="Read length that was used for sequencing. Will be used for mappability "
                                         "calculations. Should be less than the minimum fragment length. (default: 100)",
                    default=100,type=int)

parser.add_argument('--nthreads', help="Number of threads to be used. Will be slow with default value of 1", default=1,
                    type=int)
parser.add_argument('--maxmem', help="Maximum RAM (in Gigabyte) to be used by H2O for ML. Needs to be quite high."
                                     "Format: <int>G (default: 100G)", default="100G")

parser.add_argument('--nroffragmentsdrawn', help="Draw this many fragments to estimate insert size distribution. "
                                                 "(default: 200)",
                    default=200,type=int)

parser.add_argument('--polynomialdegree', help="Polynomial degree of predictors that is generated as input for a GLM."
                                               "Higher value means potentially better fitting, but longer runtime."
                                               "At the moment, only 1 and 2 are recommended. 1 (default: 1 ,i.e. no generation "
                                               "of additional variables)", default=1,type=int)
parser.add_argument('--classifier', help="Use Random Forest (\"RF\")or General Linearized Model (\"GLM\")for bias "
                                         "correction (default: RF)",
                    default="RF", choices=["GLM","RF"])

parser.add_argument('--removeoutliers', help="If set, remove coverage-outliers (Q3+1.5*IQR) from the training set",
                    action="store_true")

testing_args = parser.add_argument_group('Testing and debugging: named arguments')

testing_args.add_argument('--dontslide', help="If set, do not use the fragment-sliding method, and simply calculate the"
                    "bias factors within the bin, ignoring the flanking regions and giving equal weight to every position"
                                        " in the bin. This"
                    "flag is mainly intended for validation experiments.", action="store_true")

testing_args.add_argument('--usefixedfragsize', help="If set to an int, do not use the fragment-size distribution for the "
                                               "sliding method, but rather use the provided fixed fragment lenght."
                                               " This flag is mainly intended for validation "
                                               "experiments. (default: None)",
                        type=int, default=None )

testing_args.add_argument('--use_these_as_predictors_only', nargs='+', help='Only use these columns as predictors when '
                                                                            'using the ML classifier.'
                                                                            'Not recommended with --classifier RF.')

training=parser.add_mutually_exclusive_group()
training.add_argument('--only_calc_training_table', help="If set, only calculate a training table and then exit."
                                                         "This option can be used, e.g.,  on \"twin\" regions",
                      action="store_true")
training.add_argument('--use_provided_training_table', help="Use this training table instead of training on the same "
                                                            "regions as specified in --bedpath. This can be used if"
                                                            "a training table of twin regions is available."
                                                            "IMPORTANT: This table has to come from the same "
                                                            "sample/patient as the bigWig file is from, otherwise the"
                                                            "bias correction makes no sense.", default=None)

args = parser.parse_args()
samplename=args.samplename
bedpath=args.bedpath
binsize=args.binsize
extend_to=args.extendto
bigwigpath=args.bigwigpath
fraglen_histo=args.fraglenpath
cnvfile=args.cnvpath
correct_for_cnv=not args.dontcorrectCNVs
only_calc_training_table=args.only_calc_training_table
use_provided_training_table=args.use_provided_training_table
hg38=args.hg38path
chrlenghtspath=args.chromsizespath
mappabilitypath_fwd=args.mapppath
dontslide=args.dontslide
usefixedfragsize=args.usefixedfragsize
use_these_as_predictors_only=args.use_these_as_predictors_only
nproc=args.nthreads
maxmem=args.maxmem
draw_this_many_fragments_per_pos=args.nroffragmentsdrawn
readlength=args.readlength

remove_outliers_in_training_set=args.removeoutliers

degree=args.polynomialdegree
if degree==1 or degree==0:
    make_polynomial=False
else:
    make_polynomial=True

if args.classifier=="RF":
    use_random_forest=True
if args.classifier=="GLM":
    use_random_forest=False


################################################ INTERNAL PARAMETERS ######################################

# this is set to reduce the size of the dataframes a bit. Bias factors will be rounded to that many digits
nr_digits_to_round=6
# this can be changed if a pre-calculated mastertable should be used for training.
recalculate_mastertables=True
rebuild_model=True
# legacy:
plot=False

average_over_regions_using_list=["mean","median"]
kmeans=[2]
################################################ CODE ############################################

#### FUNCTIONS ####

# the following function is from: https://github.com/PeterUlz/TranscriptionFactorProfiling/blob/master/Scripts/analyze_coverage_around_position_log2_pysam.py
#todo: this ignores the fact that a bin could extend over the end of the CNV region because these CNV regions are usually much bigger than a bin
# TODO: this log2 ratio already comes from ichor's GC and mapp correction
def get_CNV_norm_factor(position, chrom, cnv_data):
    norm = None
    for region in cnv_data:
        reg_chrom,reg_start,reg_end,reg_log2 = region
        if chrom == reg_chrom and int(position) > int(reg_start) and position < int(reg_end):
            if not reg_log2.rstrip()=="NA":
                norm = np.power(2,float(reg_log2.rstrip()))
            break
    if norm:
        return norm
    else:
        return 1

def get_frag_distro(fraglen_histo):
    fdict={}
    with open(fraglen_histo) as f:
        counter=False
        for line in f:
            if line.startswith("insert_size"):
                counter=True
                continue
            if counter and line.replace("\n",""):
                size,nr=line.split()
                fdict[int(size)]=int(nr)

        nr_of_fragments=sum(fdict.values())
    if not usefixedfragsize:
        return list(fdict.keys()),[x/nr_of_fragments for x in fdict.values()]
    else:
        return [usefixedfragsize],[1]

# for each position in the vicinity of the bin, the GC content is multiplied by the positions weight. Bias of the bin is
# the weighted mean of the value of each such position
def get_bin_bias_incl_mapp(refgenome,mapp_fwd_bw, bin_start,longest_frag, chrom,GC_weights_for_bin,total_GC_weight,
                           mapp_fwd_weights_for_bin,total_fwd_mapp_weight,mapp_rev_weights_for_bin,
                           total_rev_mapp_weight,binsize, dinucdict,trinucdict,revcompdict,readlength):
    #assert len(GC_weights_for_bin)==len(mapp_fwd_weights_for_bin)==len(mapp_rev_weights_for_bin)
    GCbias_of_bin=0
    fwd_mappbias_of_bin=0
    rev_mappbias_of_bin=0
    ncount=0
    dinuc_res_dict=dinucdict.copy()
    trinuc_res_dict=trinucdict.copy()

    # Note: pyfaidx is indexed zero-based, as are bedfiles. PyBigWig is also zero-based
    seq=refgenome[chrom][bin_start-longest_frag:bin_start+binsize+longest_frag]
    mapp_fwd=mapp_fwd_bw.values(chrom,bin_start-longest_frag,bin_start+binsize+longest_frag)
    mapp_rev=mapp_fwd_bw.values(chrom,bin_start-longest_frag-readlength,bin_start+binsize+longest_frag-readlength)
    for pos,(GC_weight,fwd_mapp_weight,rev_mapp_weight) in enumerate(zip(GC_weights_for_bin,mapp_fwd_weights_for_bin,
                                                                         mapp_rev_weights_for_bin)):

        ## GC part
        nuc=seq[pos]
        if nuc=="G" or nuc=="g" or nuc=="c" or nuc=="C":
            GC_at_pos=1
        elif nuc=="A" or nuc=="a" or nuc=="t" or nuc=="T":
            GC_at_pos=0
        if nuc=="N" or nuc=="n":
            GC_at_pos=0.461 #0.461 is the genome-wide GC content mean
            ncount+=1

        if ncount>(longest_frag/10):
            print("Warning: a bin had", ncount, "N's. N's are counted as having 0.461 GC content.")

        #calculate the mean
        GCbias_of_bin+=GC_at_pos*GC_weight

        # trinuc part
        trinuc=seq[pos:pos+3].upper()
        try:
            trinuc_res_dict[trinuc]+=(GC_weight+GC_weights_for_bin[pos+1]+GC_weights_for_bin[pos+2])/3
        except KeyError:
            pass
            try:
                trinuc_revcomp=revcompdict[trinuc]
                trinuc_res_dict[trinuc_revcomp]+=(GC_weight+GC_weights_for_bin[pos+1]+GC_weights_for_bin[pos+2])/3
            # this can happen if we are looking at the last or second to last position
            except KeyError:
                assert len(trinuc)<3 or "N" in trinuc or "n" in trinuc
                pass

        # dinuc_part
        dinuc=trinuc[:2]
        # weight is the average of weights of the two respective positions
        try:
            dinuc_res_dict[dinuc]+=(GC_weight+GC_weights_for_bin[pos+1])/2
        # use the reverse complement if dinuc is not in dictionary
        except KeyError:
            try:
                dinuc=revcompdict[dinuc]
                dinuc_res_dict[dinuc]+=(GC_weight+GC_weights_for_bin[pos+1])/2
                # this will happen if we are looking at the last position. then dinuc is only one char and cannot be
                #found in the dict. That's ok, then we just skip that value.
            except KeyError:
                assert len(dinuc)<2 or "N" in dinuc or "n" in dinuc
                pass


        ## mappability part
        fwd_mappbias_of_bin+=mapp_fwd[pos]*fwd_mapp_weight
        rev_mappbias_of_bin+=mapp_rev[pos]*rev_mapp_weight

    GCbias_of_bin=GCbias_of_bin/total_GC_weight
    fwd_mappbias_of_bin=fwd_mappbias_of_bin/total_fwd_mapp_weight
    rev_mappbias_of_bin=rev_mappbias_of_bin/total_rev_mapp_weight
    dinuc_res_dict={key:value/total_GC_weight for key, value in dinuc_res_dict.items()}
    trinuc_res_dict={key:value/total_GC_weight for key, value in trinuc_res_dict.items()}
    return GCbias_of_bin,fwd_mappbias_of_bin,rev_mappbias_of_bin, dinuc_res_dict,trinuc_res_dict

# calculate weights combined for all "centers" using different positions in the bin as the center
def get_GC_weights_binwide(binsize,weights_GC_binsize1):
    # needs to be binsize longer than the weights if center is just a single position wide
    GC_weights_for_bin=[ [] for i in range (len(weights_GC_binsize1)+binsize)]
    # run through all postitions ("centers") within the bin
    for center in range(binsize):
        # the leftmost center just appends its weights to the list, the next appends them shifted by 1 pos and so on
        for pos in range(len(weights_GC_binsize1)):
            GC_weights_for_bin[center+pos].append(weights_GC_binsize1[pos])
        # to make the calculation of the mean meaningful, zeros need to be appended at positions where a center is not reached
        # by any fragment
        # do this for the positions that are too far downstream
        for pos in range(center):
            GC_weights_for_bin[pos].append(0)
        # and too far upstream
        for pos in range(binsize-center):
            GC_weights_for_bin[len(GC_weights_for_bin)-1-pos].append(0)
    GC_weights_for_bin=[np.mean(x) for x in GC_weights_for_bin]
    if not dontslide:
        return GC_weights_for_bin
    else:
        return [0 for i in range(longest_frag)]+[1 for i in range(binsize)]+[0 for i in range(longest_frag)]

def get_mapp_weights_binwide(binsize,frags):
    longest_frag=max(frags)
    weights_fwd={key:0 for key in range(-longest_frag,longest_frag+binsize)}
    weights_rev={key:0 for key in range(-longest_frag,longest_frag+binsize)}
    for pos in range(-longest_frag,binsize):
        first_binpos=0
        last_binpos=binsize-1
        for fraglen in frags:
            if pos <=first_binpos:
                #overlap of the fragment with the bin
                weight=max(min(binsize,fraglen-abs(pos)),0)
                weights_fwd[pos]+=weight
                weights_rev[pos+fraglen-1]+=weight
            if pos > first_binpos and pos <= last_binpos:
                weight=max(min(binsize-abs(pos), fraglen),0)
                weights_fwd[pos]+=weight
                weights_rev[pos+fraglen-1]+=weight
    if not dontslide:
        return list(weights_fwd.values()),list(weights_rev.values())
    else:

        return [0 for i in range(longest_frag)]+[1 for i in range(binsize)]+[0 for i in range(longest_frag)],\
               [0 for i in range(longest_frag)]+[1 for i in range(binsize)]+[0 for i in range(longest_frag)]

def region_worker(region, count, hg38, chromlenghts, cnvfile,frags, longest_frag,weights_GC_binsize1,GC_weights_for_standard_binsize, total_GC_weight_for_standard_binsize,
                  fwd_mapp_weights_for_standard_binsize, total_fwd_mapp_weight_for_standard_binsize,
                  rev_mapp_weights_for_standard_binsize,total_rev_mapp_weight_for_standard_binsize, binsize,bigwigpath,
                  mappabilitypath_fwd, dinucdict,trinucdict,revcompdict, readlength,extend_to):

    print("Running on region nr.", count)
    before=datetime.datetime.now()

    chrom, start, end = region.split()
    start=int(start)
    end=int(end)
    # ignore chromosomes that differ in occurence btwn the sexes and skip regions that are too close to the
    # borders of the chromosome (this would complicate things)
    if chrom=="chrX" or chrom=="chrY" or chrom not in chromlenghts.keys() or start-extend_to<0 or end+extend_to>chromlenghts[chrom]:
        print("Region on X, Y or unknown chromosome or too close to the borders. Skipping.")
        return
    # bigwig stores coverage values
    bw=pyBigWig.open(bigwigpath,"r")
    # and mappability values
    mapp_fwd_bw=pyBigWig.open(mappabilitypath_fwd,"r")
    refgenome=Fasta(hg38, as_raw=True)
    completeregion=refgenome[chrom][start-extend_to:end+extend_to]
    n_count_completeregion=completeregion.count("N")+completeregion.count("n")
    if n_count_completeregion > len(completeregion)/10:
        print("Extended region had more than 10% N's. Skipping.")
        return
    if n_count_completeregion:
        print("Warning: Extended region had", n_count_completeregion, " N's in the reference sequence. Not skipping.")

    if correct_for_cnv:
        cnvdata=[]
        with open(cnvfile,"r") as cnvf:
            cnvdata=cnvf.readlines()
        cnvdata=[region.split("\t") for region in cnvdata]
    # devide the central peak into 5 bins, such that ROIs with peaks of different width have the same number of bins
    # and are thus more comparable
    regiondict_peak=OrderedDict()
    width=end-start
    regiondict_peak["0to10"]={"coverage":np.mean(bw.values(chrom,start,int(start+0.1*width)))}
    regiondict_peak["0to10"]["coord"]=(chrom,start,int(start+0.1*width))
    regiondict_peak["10to25"]={"coverage":np.mean(bw.values(chrom,int(start+0.1*width),int(start+0.25*width)))}
    regiondict_peak["10to25"]["coord"]=(chrom,int(start+0.1*width),int(start+0.25*width))
    regiondict_peak["25to75"]={"coverage":np.mean(bw.values(chrom,int(start+0.25*width),int(start+0.75*width)))}
    regiondict_peak["25to75"]["coord"]=(chrom,int(start+0.25*width),int(start+0.75*width))
    regiondict_peak["75to90"]={"coverage":np.mean(bw.values(chrom,int(start+0.75*width),int(start+0.90*width)))}
    regiondict_peak["75to90"]["coord"]=(chrom,int(start+0.75*width),int(start+0.9*width))
    regiondict_peak["90to100"]={"coverage":np.mean(bw.values(chrom,int(start+0.90*width),end))}
    regiondict_peak["90to100"]["coord"]=(chrom,int(start+0.9*width),end)
    # note that this implementation assumes that the end coordinate of the ROI should not be part of the peak
    # anymore (python interprets the end cooridinates in ranges etc as the first coordinates that should not be part
    # of a sequence etc. anymore)

    # add GC information for these bins with specific lenghts:
    for bin_name, bin_dict in regiondict_peak.items():
        bin_start=bin_dict["coord"][1]
        this_binsize=bin_dict["coord"][2]-bin_dict["coord"][1]
        if correct_for_cnv:
            norm_factor=get_CNV_norm_factor(bin_start,bin_dict["coord"][0],cnvdata)
            bin_dict["coverage"]=round(bin_dict["coverage"]/norm_factor,nr_digits_to_round)
        fwd_mapp_weights, rev_mapp_weights=get_mapp_weights_binwide(this_binsize,frags)
        GC_weights=get_GC_weights_binwide(this_binsize,weights_GC_binsize1)
        total_GC_weight=sum(GC_weights)
        total_fwd_mapp_weight=sum(fwd_mapp_weights)
        total_rev_mapp_weight=sum(rev_mapp_weights)
        GC,mappfwd,mapprev, dinucresultdict,\
        trinucresultdict=get_bin_bias_incl_mapp(refgenome,mapp_fwd_bw, bin_start,longest_frag, chrom,GC_weights,total_GC_weight,
                                                  fwd_mapp_weights,total_fwd_mapp_weight, rev_mapp_weights,
                                                  total_rev_mapp_weight,this_binsize, dinucdict,trinucdict,revcompdict,readlength)
        # the GC bin_dict of the bin is the mean of all positions in the bin
        bin_dict["GCcont"]=round(GC,nr_digits_to_round)
        bin_dict["mapp_fwd"]=round(mappfwd,nr_digits_to_round)
        bin_dict["mapp_rev"]=round(mapprev,nr_digits_to_round)
        bin_dict["mapp_max"]=round(max(mappfwd,mapprev),nr_digits_to_round)
        for key, value in dinucresultdict.items():
            bin_dict[key]=round(value,nr_digits_to_round)
        for key, value in trinucresultdict.items():
            bin_dict[key]=round(value,nr_digits_to_round)



    regiondict_left=OrderedDict()
    # This treats the bins left of the center. Devide the data into bins and associate the mean coverage value
    # of all its positions with a bin
    posnr=1
    # bin furthest to the left is called is called bin -<nr of bins left of center>
    binnr=int(-extend_to/binsize)
    for value in range(extend_to):
        if posnr==binsize:
            posnr=0
            # e.g. start=1000 and binnr=-5 and binsize=10, then bin -5 should stretch from 950 to 960
            binstart=start+(binnr*binsize)
            binend=start+(binnr*binsize)+binsize
            if correct_for_cnv:
                norm_factor=get_CNV_norm_factor(binstart,chrom,cnvdata)
            else:
                norm_factor=1
                                            #this gets the coverage values from the bigwig file
            regiondict_left["bin_"+str(binnr)]={"coverage":round(np.mean(bw.values(chrom,binstart,binend))/norm_factor,nr_digits_to_round)}
            regiondict_left["bin_"+str(binnr)]["coord"]=(chrom,binstart,binend)
            binnr+=1
        posnr+=1

    # same as above but for bins right of center
    # bin closest to center is called bin 0
    regiondict_right=OrderedDict()
    binnr=0
    posnr=1
    for value in range(extend_to):
        if posnr==binsize:
            posnr=0
            binstart=end+(binnr*binsize)
            binend=end+(binnr*binsize)+binsize
            if correct_for_cnv:
                norm_factor=get_CNV_norm_factor(binstart,chrom,cnvdata)
            else:
                norm_factor=1
            regiondict_right["bin_"+str(binnr)]={"coverage":round(np.mean(bw.values(chrom,binstart,binend))/norm_factor,nr_digits_to_round)}
            regiondict_right["bin_"+str(binnr)]["coord"]=(chrom,binstart,binend)
            binnr+=1
        posnr+=1

    for regiondict in [regiondict_left,regiondict_right]:
        # add GC information for each bin:
        for bin_name, bin_dict in regiondict.items():
            # slide a number of randomly drawn fragments over the position and get their mean GC bin_dict
            GC,mappfwd, mapprev, dinucresultdict, trinucresultdict=get_bin_bias_incl_mapp(refgenome, mapp_fwd_bw, bin_dict["coord"][1],longest_frag,chrom,GC_weights_for_standard_binsize,total_GC_weight_for_standard_binsize,
                                                       fwd_mapp_weights_for_standard_binsize, total_fwd_mapp_weight_for_standard_binsize,
                                                       rev_mapp_weights_for_standard_binsize,total_rev_mapp_weight_for_standard_binsize,
                                                       binsize,dinucdict,trinucdict,revcompdict,readlength)
            # the GC bin_dict of the bin is the mean of all positions in the bin
            bin_dict["GCcont"]=round(GC,nr_digits_to_round)
            bin_dict["mapp_fwd"]=round(mappfwd,nr_digits_to_round)
            bin_dict["mapp_rev"]=round(mapprev,nr_digits_to_round)
            bin_dict["mapp_max"]=round(max(mappfwd,mapprev),nr_digits_to_round)
            for key, value in dinucresultdict.items():
                bin_dict[key]=round(value,nr_digits_to_round)
            for key, value in trinucresultdict.items():
                bin_dict[key]=round(value,nr_digits_to_round)

    # this stores the bins in the correct order, from upstream to downstream
    regiondict_all=regiondict_left
    regiondict_all.update(regiondict_peak)
    regiondict_all.update(regiondict_right)

    print("this took",datetime.datetime.now()-before)
    return (chrom+":"+str(start)+"-"+str(end),regiondict_all)



############### MAIN ############
if __name__ == '__main__':
    # write log file with parameters
    subprocess.check_call("mkdir -p "+samplename,shell=True)
    with open(samplename+"/"+str(datetime.datetime.now()).replace(" ","")+".log","w") as logf:
        for arg, value in sorted(vars(args).items()):
            print("Argument %s: %s"%(arg,value),file=logf)

    if recalculate_mastertables:
        ## preparations ##
        #output
        subprocess.check_call("mkdir -p "+samplename,shell=True)

        #binsize
        assert not extend_to % binsize
        chromlenghts = {}
        with open(chrlenghtspath) as chromlenghtsfile:
            for line in chromlenghtsfile:
                chrom, length = line.split()
                chromlenghts[chrom] = int(length)

        # prepare a representative subsample of fragment lengths
        fraglengths,fragweights=get_frag_distro(fraglen_histo)
        random.seed(a=42)
        frags=random.choices(fraglengths,fragweights,k=draw_this_many_fragments_per_pos)
        longest_frag=max(frags)
        print("The maximum fragment length: ",longest_frag)

   
        ### Weight calculations

        # calculate weights for the first position of the bin
        # count how many frags are longer or equal to this many bases from the center
        # GC content of a position is more important the closer the position is to the center and the more fragments extend that far
        """
        abs(pos):321 0 123
        ------------
        how often a fragment overlaps with a position as it is slided           
                 xxx|x|
                  xx|x|x
                   x|x|xx
                    |x|xxx
                   y|y|
                    |y|y       
        
        """
        # GC weights for a bin with width=1
        # This is faster than adopting the weightcalculation of mappability for GC and produces exactly the same results
        # (verified)
        weights_GC_binsize1=[]
        for pos in range(-longest_frag,longest_frag):
            weight=sum([fraglen-abs(pos) for fraglen in frags if fraglen>=abs(pos)])
            weights_GC_binsize1.append(weight)

        # GC weights for a bin with standard size
        GC_weights_for_standard_binsize=get_GC_weights_binwide(binsize,weights_GC_binsize1)
        total_GC_weight_for_standard_binsize=sum(GC_weights_for_standard_binsize)
        # mappability weights (forward and reverse) for a bin with standard size
        fwd_mapp_weights_for_standard_binsize,rev_mapp_weights_for_standard_binsize=get_mapp_weights_binwide(binsize,frags)
        total_fwd_mapp_weight_for_standard_binsize=sum(fwd_mapp_weights_for_standard_binsize)
        total_rev_mapp_weight_for_standard_binsize=sum(rev_mapp_weights_for_standard_binsize)

        # dictionary of all di- and trinucleotides:
        nucdict={}
        dinucdict={}
        trinucdict={}

        for nuc in ["A","T","G","C"]:
            nucdict[nuc]=0

        k=list(nucdict.keys())
        for nuc in ["A","T","G","C"]:
            for key in k:
                dinuc=key+nuc
                dinuc_seq=Seq.Seq(dinuc)
                if not str(dinuc_seq.reverse_complement()) in dinucdict.keys():
                    dinucdict[dinuc]=0
                nucdict[dinuc]=0

        k=list(nucdict.keys())
        for nuc in ["A","T","G","C"]:
            for key in k:
                if len(key)==2:
                    trinuc=key+nuc
                    trinuc_seq=Seq.Seq(trinuc)
                    if not str(trinuc_seq.reverse_complement()) in trinucdict.keys():
                        trinucdict[trinuc]=0
                    nucdict[trinuc]=0
        nucdict={k:v for k,v in nucdict.items() if len(k)>1}
        # better build this here, reverse_complement is quite slow
        revcompdict={key:str(Seq.Seq(key).reverse_complement()) for key in list(nucdict.keys())}


        # list of regions present in the bed file
        regions=[]
        with open(bedpath) as bed:
            regions=bed.readlines()

        mt=dict()
        # Run the rest in a multiprocessed fashion. The starmap passes the arguments to the worker, on tuple of arguments
        #at a time. results containes an iterator over (region, regiondict) tuples when returned.
        pool=multiprocessing.Pool(processes=nproc)
        with pool as mypool:
            results=pool.starmap(region_worker,((region,count, hg38, chromlenghts, cnvfile,
                                                 frags, longest_frag,weights_GC_binsize1,
                                                 GC_weights_for_standard_binsize,total_GC_weight_for_standard_binsize,
                                                 fwd_mapp_weights_for_standard_binsize,
                                                 total_fwd_mapp_weight_for_standard_binsize,
                                                 rev_mapp_weights_for_standard_binsize,
                                              total_rev_mapp_weight_for_standard_binsize, binsize,bigwigpath,
                                                 mappabilitypath_fwd, dinucdict,trinucdict,revcompdict,readlength,extend_to) for count, region in enumerate(regions) ))

        print("region-wise calculation complete")
        before=datetime.datetime.now()
        # containes None if region is "rejected" by the worker
        mt=dict(filter(None, results))

        # data frame of the master table
        print("Turning the dict into a pandas dataframe. This might take a while...")
        before=datetime.datetime.now()
        # the following is nessesary to get the bins in the master table into the right order
        mdf=pd.DataFrame.from_dict({(region,bin):mt[region][bin] for region in mt.keys() for bin in mt[region].keys()},orient='columns')
        mdf=mdf.reindex(labels=[(region,bin) for region in mt.keys() for bin in mt[region].keys()],axis="columns").transpose()
        mdf.index.names=["region", "bin"]
        print("Done. This took",datetime.datetime.now()-before)
        print("Size of the dataframe (bytes): ",sys.getsizeof(mdf))
        print("This is what the mastertable looks like: \n Statistics:\n", mdf.describe())
        print("And the first couple of rows: \n",mdf.head())
        before=datetime.datetime.now()
        if not only_calc_training_table:
            print("Saving pandas dataframe to csv. This might take a while...")
            mdf.to_csv(samplename+"/master_table.csv")
            print("Done. This took",datetime.datetime.now()-before)
        if not use_provided_training_table:
            #this excludes peak regions
            training_df=mdf.drop(labels=["0to10","10to25","25to75","75to90","90to100"],axis="index", level="bin")
            training_df.to_csv(samplename+"/training_table.csv")

        if only_calc_training_table:
            print("Calculation of training table complete. Exiting.")
            sys.exit()
    ################################# H20 ML part ############################################

    subprocess.check_call("mkdir -p "+samplename+"/tmp",shell=True)
    os.environ["TMPDIR"] = os.getcwd()+"/"+samplename+"/tmp"
    tmpdir=str(os.getcwd()+"/"+samplename+"/tmp")
    h2o.init(max_mem_size=maxmem, nthreads=nproc, ice_root=tmpdir)

    response_col="coverage"

    if rebuild_model:
        if not use_provided_training_table:
            training_df=pd.read_csv(samplename+"/training_table.csv")
        else:
            training_df=pd.read_csv(use_provided_training_table)
        # first 2 columns need to be renamed
        training_df.columns=["region","bin"]+list(training_df.columns)[2:]
        needed_cols=list(training_df.columns.values)
        needed_cols.remove("bin")
        needed_cols.remove("region")
        needed_cols.remove("coord")
        training_df_nonindexed=training_df[needed_cols]
        if not use_these_as_predictors_only:
            predictors=needed_cols.copy()
            predictors.remove("coverage")
        else:
            predictors=use_these_as_predictors_only

        print("These are the non-polynomial predictors", predictors)
        predictors_unpolyed=predictors.copy()

        ## make all predictors polynomial
        if make_polynomial:
            make_poly_of_this=training_df_nonindexed[predictors]
            poly=PolynomialFeatures(degree=degree)
            poly=poly.fit(make_poly_of_this.values)
            polyed=poly.transform(make_poly_of_this.values)
            polyed_names=poly.get_feature_names(predictors)
            polyed=pd.DataFrame(polyed)
            polyed.columns=polyed_names
            polyed[response_col]=training_df_nonindexed[response_col].values
            training_df_nonindexed=polyed
            predictors=polyed_names
            predictors.remove("1")

            print("These are the polynomial predictors",predictors)
            pred_and_resp=predictors+[response_col]
            before=datetime.datetime.now()
            print("INFO ABOUT THE DF - unconverted:")
            print(training_df_nonindexed[pred_and_resp].info(memory_usage='deep'))
            print("Creating an H2o dataframe - this may take a while...")
            h20data=h2o.H2OFrame(training_df_nonindexed[pred_and_resp])
            print(h20data["coverage"].describe())
            print("Done. This took",datetime.datetime.now()-before, "Working with this df (head):\n", h20data.head())


        else:
            print("You chose not to make polynomials for the linear model")
            pred_and_resp=predictors+[response_col]
            before=datetime.datetime.now()
            print("Creating a H2o dataframe - this may take a while...")
            h20data=h2o.H2OFrame(training_df_nonindexed[pred_and_resp])
            print("Done. This took",datetime.datetime.now()-before, "Working with this df (head):\n", h20data.head())

        train, test = h20data.split_frame(ratios=[.8],seed=3)

        print("Rows in training data before removing outliers:", train.dim[0])
        print("Max values train before outlier removal:", train["coverage"].max())
        if remove_outliers_in_training_set:
            traindf=train.as_data_frame()
            # REMOVE OUTLIERS: remove those that have higher coverage q3+1.5*IQR
            Q3=np.percentile(traindf["coverage"],75)
            outlier_treshold=Q3+1.5*(Q3-np.percentile(traindf["coverage"],25))
            print("The treshold for removing outliers:", outlier_treshold)
            df_no_outlier=traindf[traindf["coverage"]< outlier_treshold]
            train=h2o.H2OFrame(df_no_outlier)
            print("Rows in training data after removing outliers:", train.dim[0])
            print("Max values train:", train["coverage"].max())
            print("Max values test:", test["coverage"].max())
        else:
            print("you chose not to remove outliers in the training set.")


        if plot:
            print("The following statistics refer to the complete master-dataset:")
            slope, intercept, r_value, p_value, std_err = stats.linregress(training_df_nonindexed["GCcont"],training_df_nonindexed["coverage"])
            line = slope*training_df_nonindexed["GCcont"]+intercept
            plt.plot(training_df_nonindexed["GCcont"],training_df_nonindexed["coverage"],"kx",training_df_nonindexed["GCcont"],line)
            print("GCcont","r2val",r_value**2,"pval",p_value)
            plt.show()

            slope, intercept, r_value, p_value, std_err = stats.linregress(training_df_nonindexed["mapp_max"],training_df_nonindexed["coverage"])
            line = slope*training_df_nonindexed["mapp_max"]+intercept
            plt.plot(training_df_nonindexed["mapp_max"],training_df_nonindexed["coverage"],"kx",training_df_nonindexed["mapp_max"],line)
            print("mapp_max","r2val",r_value**2,"pval",p_value)
            plt.show()

            slope, intercept, r_value, p_value, std_err = stats.linregress(training_df_nonindexed["mapp_rev"],training_df_nonindexed["coverage"])
            line = slope*training_df_nonindexed["mapp_rev"]+intercept
            plt.plot(training_df_nonindexed["mapp_rev"],training_df_nonindexed["coverage"],"kx",training_df_nonindexed["mapp_rev"],line)
            print("mapp_rev","r2val",r_value**2,"pval",p_value)
            plt.show()

            slope, intercept, r_value, p_value, std_err = stats.linregress(training_df_nonindexed["mapp_fwd"],training_df_nonindexed["coverage"])
            line = slope*training_df_nonindexed["mapp_fwd"]+intercept
            plt.plot(training_df_nonindexed["mapp_fwd"],training_df_nonindexed["coverage"],"kx",training_df_nonindexed["mapp_fwd"],line)
            print("mapp_fwd","r2val",r_value**2,"pval",p_value)
            plt.show()


        folds=2
        #alpha==1 -> LASSO (forces to 0); alpha==0 -> ridge
        myalpha = 1
        mylambda = 0.005
        #compute_p_values=True,remove_collinear_columns=True

        if not use_random_forest:
            mymodel = H2OGeneralizedLinearEstimator(family="gaussian", alpha=myalpha, lambda_search=True ,nfolds=folds, missing_values_handling="skip")#, interactions=predictors)#, interaction_pairs=pairs) #lambda_=mylambda
        if use_random_forest:
            mymodel = H2ORandomForestEstimator(ntrees=50, seed=42, nfolds=folds)
        
        mymodel.train(x=predictors, y=response_col, training_frame=train, validation_frame=test)#, training_frame=h20data)
        print(mymodel)
        if use_random_forest:
            mymodel._model_json['output']['variable_importances'].as_data_frame().to_csv(samplename+"/RF_coefficients.csv")
            if plot:
                mymodel.varimp_plot()
        else:
            mymodel._model_json['output']['coefficients_table'].as_data_frame().to_csv(samplename+"/GLM_coefficients.csv")
            if plot:
                mymodel.std_coef_plot()

    else:
        modelpath=glob.glob(os.getcwd()+"/"+samplename+"/*model*")[-1]
        print("Loading model ",modelpath)
        mymodel=h2o.load_model(modelpath)


    # use the model to make predcitions:
    print("Running predictions with the model")
    complete_df=pd.read_csv(samplename+"/master_table.csv")
    # first two columns are region and bin
    complete_df.columns=["region","bin"]+list(complete_df.columns)[2:]
    if not use_these_as_predictors_only:
        predictors=list(complete_df.columns.values)
        predictors.remove("coverage")
        predictors.remove("bin")
        predictors.remove("region")
        predictors.remove("coord")
    else:
        predictors=use_these_as_predictors_only
    predictors_unpolyed=predictors.copy()

    if make_polynomial:
        make_poly_of_this=complete_df[predictors_unpolyed]
        poly=PolynomialFeatures(degree=degree)
        poly=poly.fit(make_poly_of_this.values)
        polyed=poly.transform(make_poly_of_this.values)
        polyed_names=poly.get_feature_names(predictors_unpolyed)
        polyed=pd.DataFrame(polyed)
        polyed.columns=polyed_names
        polyed[response_col]=complete_df[response_col].values
        complete_df=polyed
        predictors=polyed_names
        predictors.remove("1")

    pred_and_resp=predictors+[response_col]
    before=datetime.datetime.now()
    # todo: this could be circumvented if a model has been built anyway
    print("Creating an H2o dataframe from mastertable - this may take a while...")
    masterdf_h2o=h2o.H2OFrame(complete_df[pred_and_resp])
    print("Done. This took", datetime.datetime.now()-before)

    masterdf_h2o_cov_pred=mymodel.predict(masterdf_h2o)

    complete_df=pd.read_csv(samplename+"/master_table.csv")
    complete_df["pred_coverage"]=(masterdf_h2o_cov_pred.as_data_frame()).values
    complete_df.columns=["region","bin"]+list(complete_df.columns)[2:]
    complete_df["corr_coverage"]=(complete_df["coverage"].subtract(complete_df["pred_coverage"])).values
    complete_df.to_csv(samplename+"/mastertable_w_corrections.csv.gz", compression="gzip")
    subprocess.check_call("rm "+samplename+"/master_table.csv",shell=True)
    subprocess.check_call("rm "+samplename+"/training_table.csv || true ",shell=True)
    print("this is what the corrected df looks like: Statistics:  \n ",complete_df.describe())
    print("And a couple of rows:")
    print(complete_df.head())

    # Quit the h2o cluster
    h2o.cluster().shutdown()
    #subprocess.check_call("rm -r "+tmpdir+" 2> /dev/null",shell=True)

    # Prepare for plotting:
    nr_left_bins=len(list(range(-extend_to,0,binsize)))
    binnamelist=["bin_"+str(x) for x in range(int(-extend_to/binsize),0)]+["0to10","10to25","25to75","75to90","90to100"]+["bin_"+str(x) for x in range(0, int(extend_to/binsize))]

    for coverage_type in ["coverage","corr_coverage"]:
        print("For",coverage_type,":")

        for average_over_regions_using in average_over_regions_using_list:
            # get median (over all regions) coverage values for each bin
            print("Using",average_over_regions_using, "to average over regions")
            if average_over_regions_using=="median":
                average_per_binnr=pd.DataFrame(complete_df.groupby(['bin'])[coverage_type].median())
            if average_over_regions_using=="mean":
                average_per_binnr=pd.DataFrame(complete_df.groupby(['bin'])[coverage_type].mean())
            average_per_binnr.columns=["averagecoverage"]
            average_per_binnr=average_per_binnr.reset_index()
            # the dataframe is not sorted properly (starts with "0to10"), need to sort
            sorthelp=average_per_binnr["bin"].apply(lambda x: {bin_entry:value for value,bin_entry in enumerate(binnamelist)}[x])
            average_per_binnr["sorthelp"]=sorthelp
            average_per_binnr=average_per_binnr.sort_values(by="sorthelp")
            average_per_binnr.to_csv(samplename+"/"+coverage_type+"_"+average_over_regions_using+"_per_bin.csv")