#!/usr/bin/env bash

# this is the directory in which the script is located:
DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" >/dev/null 2>&1 && pwd )"
# paths to other scripts are relative to DIR
TWINREGIONSCRIPT=$DIR/get_twinregion_trainingset.py
MODELSCRIPT=$DIR/LIQUORICE_v0.3.1.py

# https://github.com/nhoffman/argparse-bash
ARGPARSE_DESCRIPTION="LIQUORICE: A tool for bias correction in cfDNA WGS datasets"
source $DIR/argparse.bash || exit 1
argparse "$@" <<EOF || exit 1

# an argparse action that converts relative paths to full paths; form: https://gist.github.com/brantfaircloth/1252339
class FullPaths(argparse.Action):
    """Expand user- and relative-paths"""
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, os.path.abspath(os.path.expanduser(values)))

required_keyword_args = parser.add_argument_group('required named arguments')

required_keyword_args.add_argument('--samplename', help='Name of the sample that is being processed. '
                                                              'This will be used for ouput plots and the names of '
                                                              'directories and log files', required=True)
required_keyword_args.add_argument('--bedpath', help='Path to the BED file that contains the regions that '
                                                           'should be analyzed. Should be in hg38. '
                                                           'Should be in the format chromName\\tstart\\tend',
                                   required=True,action = FullPaths)
required_keyword_args.add_argument('--bigwigpath', help='Path to the BigWig file that contains the coverage information'
                                                        ' of the sample. Should probably have a small binsize, '
                                                        'we are working with bamCoverage with bs=1'
                                                        '', required=True,action = FullPaths)
required_keyword_args.add_argument('--fraglenpath', help='Path to the output .txt file of Picard\'s '
                                                          'CollectInsertSizeMetrics tool.' ,required=True,
                                                          action = FullPaths)
required_keyword_args.add_argument('--cnvpath', help='Path to ichorCNA\s <samplename>.correctedDepth.txt file. Required'
                                                     'for correction of CNV biases' ,required=True,action = FullPaths)

#hg38 = "/home/peter/ctDNA/data/reference_data/hg38/hg38.fa"
required_keyword_args.add_argument('--hg38path', help='Path to a hg38.fa file' ,required=True,action = FullPaths)
# todo: ship this together with the program, and do not require
#chrlenghtspath = "/home/peter/ctDNA/data/reference_data/hg38/hg38_selfmade_chrom.sizes"
required_keyword_args.add_argument('--chromsizespath', help='Path to a chrom.sizes file for hg38' ,required=True,
action = FullPaths)
#mappabilitypath_fwd="/home/peter/ctDNA/data/reference_data/hg38/hg38_mappability_75bp.bigwig"
required_keyword_args.add_argument('--mapppath', help='Path to a bigWig file that contains (forward) mappability'
                                                            ' values for every base in the reference genome. '
                                                            'Can be calculated with gem-mappability for the appropriate '
                                                            'read length' ,required=True,action = FullPaths)


parser.add_argument('--binsize', help="Bin size that is to be used for the flanking regions. All bins except"
                                           " the core region,"
                                           "specified in --bedpath (which will be splitted differently), will have "
                                           "this binsize. "
                                           "Bin size is important for the resolution of the output plots & data, and "
                                           "for the bias model itself (default: 180)", default=180, type=int)

parser.add_argument('--extendto', help="Size of the flanking region, in bp. Must be devidable by bin size "
                                       "The analyzed regions are extended by "
                                            "this value, both upstream and downstream. Example:"
                                            "core region=150 bp, extendto=1000 --> length of analyzed region in total: "
                                            "2150 bp (default: 18000) ", default=18000, type=int)
parser.add_argument('--dontcorrectCNVs', help="If set, do not correct for CNV biases. You can set --cnvpath to 'None' if"
                                              "this option is used. If a file is given, it will be ignored."
                    , action="store_true")

parser.add_argument('--readlength', help="Read length that was used for sequencing. Will be used for mappability "
                                         "calculations. Should be less than the minimum fragment length. (default: 100)",
                    default=100,type=int)

parser.add_argument('--nthreads', help="Number of threads to be used. Will be slow with default value of 1", default=1,
                    type=int)
parser.add_argument('--maxmem', help="Maximum RAM (in Gigabyte) to be used by H2O for ML. Needs to be quite high."
                                     "Format: <int>G (default: 100G)", default="100G")

parser.add_argument('--nroffragmentsdrawn', help="Draw this many fragments to estimate insert size distribution. "
                                                 "(default: 200)",
                    default=200,type=int)

parser.add_argument('--polynomialdegree', help="Polynomial degree of predictors that is generated as input for a GLM."
                                               "Higher value means potentially better fitting, but longer runtime."
                                               "At the moment, only 1 and 2 are recommended. 1 (default: 1 ,i.e. no generation "
                                               "of additional variables)", default=1,type=int)
parser.add_argument('--classifier', help="Use Random Forest (\"RF\")or General Linearized Model (\"GLM\")for bias "
                                         "correction (default: RF)",
                    default="RF", choices=["GLM","RF"])

parser.add_argument('--removeoutliers', help="If set, remove coverage-outliers (Q3+1.5*IQR) from the training set",
                    action="store_true")

testing_args = parser.add_argument_group('Testing and debugging: named arguments')

testing_args.add_argument('--dontslide', help="If set, do not use the fragment-sliding method, and simply calculate the"
                    "bias factors within the bin, ignoring the flanking regions and giving equal weight to every position"
                                        " in the bin. This"
                    "flag is mainly intended for validation experiments.", action="store_true")

testing_args.add_argument('--usefixedfragsize', help="If set to an int, do not use the fragment-size distribution for the "
                                               "sliding method, but rather use the provided fixed fragment lenght."
                                               " This flag is mainly intended for validation "
                                               "experiments. (default: None)",
                        type=int, default=None )

testing_args.add_argument('--use_these_as_predictors_only', nargs="+", help='Only use these columns as predictors when '
                                                    'using the ML classifier. Not recommended with --classifier RF.')

### NOTE: this is only relelvant in the runscript:

parser.add_argument('--outdir', help="If set, save all output to that dir instead of saving to the current directory",
action = FullPaths)

parser.add_argument('--usetwinstotrain', help="If set, do not train on the same data that should be corrected, but"
"rather generate a set of random regions with similar properties in terms of GCcont, mappability, and nr. of N\'s. Then"
"use these \"twin\" regions to generate a training set, on which the classifier is trained. May help to improve correction"
"if the signal of interest is associated with bias factors. (e.g. if a DHS is expected to have a drop in coverage"
"independently of GC content, but if GC content is also increased around the DHS. This setting increases runtime.",
                      action="store_true")

parser.add_argument('--twinsbedfile', help="Use this bed file as twin regions instead of generating a new set of"
"regions. Requires --usetwinstotrain, or will be ignored otherwise.",action = FullPaths)

EOF

if [ "$OUTDIR" ]
then
    mkdir -p $OUTDIR
    cd $OUTDIR
fi

### Pass on the flags to the python script:

additionalflags=""
if [ "$DONTCORRECTCNVS" ]
then
    additionalflags=${additionalflags}"--dontcorrectCNVs "
fi
if [ "$REMOVEOUTLIERS" ]
then
    additionalflags=${additionalflags}"--removeoutliers "
fi
if [ "$DONTSLIDE" ]
then
    additionalflags=${additionalflags}"--dontslide "
fi
if [ "$USEFIXEDFRAGSIZE" ]
then
    additionalflags=${additionalflags}"--usefixedfragsize $USEFIXEDFRAGSIZE"
fi

if [ "$USE_THESE_AS_PREDICTORS_ONLY" ]
then
    additionalflags=${additionalflags}"--use_these_as_predictors_only "
    for arg in ${USE_THESE_AS_PREDICTORS_ONLY[@]}
    do
        additionalflags=${additionalflags}" "$arg
    done
    additionalflags=${additionalflags}" "
fi

### Start the python script, either in twin mode or not
BEDNAME=$(basename $BEDPATH .bed)

# script requires that bedfile only has three columns, chrom, start, and end.
cut -f1,2,3 $BEDPATH > ${BEDNAME}.bed
BEDPATH=$(realpath ${BEDNAME}.bed)


if [ "$USETWINSTOTRAIN" ]
then
    mkdir -p twin_regions
    cd twin_regions
    if [ "$TWINSBEDFILE" ]
    then
        echo "Using provided twin bed file $TWINSBEDFILE"

    else
        echo "No twins bed file provided, finding twin regions ..."
        python3 $TWINREGIONSCRIPT --bedpath $BEDPATH --extendto $EXTENDTO --outfile twin_regions.bed --hg38path $HG38PATH \
        --chromsizespath $CHROMSIZESPATH --mapppath $MAPPPATH
        TWINSBEDFILE=twin_regions.bed
        echo "done"
    fi
    echo "Building training table with twin regions"
    python3 -u $MODELSCRIPT --samplename $SAMPLENAME --bedpath $TWINSBEDFILE \
    --bigwigpath ${BIGWIGPATH} --fraglenpath ${FRAGLENPATH} --cnvpath ${CNVPATH} --hg38path $HG38PATH \
     --chromsizespath $CHROMSIZESPATH --mapppath $MAPPPATH --binsize ${BINSIZE} \
     --extendto $EXTENDTO --readlength $READLENGTH --nthreads $NTHREADS --maxmem $MAXMEM \
     --nroffragmentsdrawn $NROFFRAGMENTSDRAWN --polynomialdegree $POLYNOMIALDEGREE --classifier $CLASSIFIER \
     --only_calc_training_table $additionalflags > ${SAMPLENAME}_twin_regions.log

     #just to make clear that the training table is made from twins:
     mv ${SAMPLENAME}/training_table.csv ${SAMPLENAME}/twin_training_table.csv
     echo "done"

    cd ..
    echo "Using twin training table to correct regions of interest"
    additionalflags=${additionalflags}"--use_provided_training_table twin_regions/${SAMPLENAME}/twin_training_table.csv"
else
    echo "Using original regions training data to correct original regions"
fi

    python3 -u $MODELSCRIPT --samplename $SAMPLENAME --bedpath $BEDPATH \
    --bigwigpath ${BIGWIGPATH} --fraglenpath ${FRAGLENPATH} --cnvpath ${CNVPATH} --hg38path $HG38PATH \
     --chromsizespath $CHROMSIZESPATH --mapppath $MAPPPATH --binsize ${BINSIZE} \
     --extendto $EXTENDTO --readlength $READLENGTH --nthreads $NTHREADS --maxmem $MAXMEM \
     --nroffragmentsdrawn $NROFFRAGMENTSDRAWN --polynomialdegree $POLYNOMIALDEGREE --classifier $CLASSIFIER \
     $additionalflags > ${SAMPLENAME}.log
     rm twin_regions/${SAMPLENAME}/twin_training_table.csv &>/dev/null #to ignore errors if file does not exist
     echo "done"

rm ${BEDNAME}.bed